import os
from prompt_toolkit import PromptSession
from .misc import set_env

def open_in_editor(prompt_session=None, extension='.txt', content=None):
  if not prompt_session:
    prompt_session = PromptSession(erase_when_done=True)

  with set_env('EDITOR', os.environ.get('EDITOR', 'vim')):
    buffer = prompt_session.layout.current_buffer
    buffer.text = content or ''
    buffer.open_in_editor(validate_and_handle=True)
    text = prompt_session.prompt(
        buffer.text,
        enable_open_in_editor=True,
        tempfile_suffix=extension,
    )

  return text



class CommandManager:

  registry = {}

  def register(self, name):
    def decorator(cls):
      cls.name = name
      self.registry[name] = cls
      return cls
    return decorator

  def exec(self, repl, text, print_text):
    text = text.strip()

    if text == '':
      return ('break', text)

    for Command in self.get_command_list():
      cmd = Command(repl, text, print_text)
      if cmd.matches():
        return cmd.exec()

    print_text(text)
    print()

    return ('continue', text)

  def get_command_list(self):
    command_list = self.registry.values()
    return sorted(command_list, key=lambda c: c.sort_index)

  def bind_keys(self, kb):
    for Command in self.get_command_list():
      shortcut = Command.shortcut
      if not shortcut or len(shortcut) == 0:
        continue
      shortcut = [ s.lower() for s in shortcut ]
      self.bind_key(kb, Command.name, shortcut)

  def bind_key(self, kb, cmd_name, shortcut):
    @kb.add(*shortcut)
    def _(event):
      buffer = event.current_buffer
      buffer.text = cmd_name
      buffer.validate_and_handle()

Commands = CommandManager()



import math, uuid, re

from .base_mode import BaseMode, register_mode
from ..llms.gpt3 import GPT3

@register_mode('synth-chat')
class SynthChatMode(BaseMode):

  title = 'ChattyGPT'

  line_sep = '---------'
  line_sep_token = '45537'

  def __init__(self, state={}):
    self.model = 'text-davinci-003'
    self.temperature = 0.7
    self.llm = GPT3()

    self.profile = 'assistant'
    self.seed = ''
    self.response_prefix = ''

    self.human_name = 'Eden'
    self.authority_name = 'Boss'
    self.persona_name = 'Delphi'
    self.persona_bio = ''

    self.pinned_summary = lambda: f'{self.human_name} demanded I give comprehensive answers, including detailed code, writing, guides, and more. I should use Markdown formatting (lists like "1.", headers like "# title", code blocks like ```js etc). When asked a question that is nonsense, trickery, or has no clear answer, I must respond explaining what\'s wrong with the question. '
    self.prologue = [
      # {
      #   'source': 'server',
      #   # 'text': f"Hello! How can I help? I'm going to write some awesome code for you!",
      #   # 'text': f"Hello! How can I help? I'll tell you if I don't know something.",
      # },
    ]

    self.summary_header = lambda: f'!! {self.persona_name}\'s Old Live Chat Notes:'
    self.conversation_header = lambda: f'!! Recent Live Chat Between {self.human_name} and {self.persona_name}:'

    self.max_summaries = 8
    self.soft_max_depth = 18
    self.min_rollup_tokens = 200

    self.max_response_tokens = 1000
    self.max_prompt_tokens = 4000

    self.soft_max_message_tokens = 150
    self.soft_max_prompt_tokens = min(2000, self.max_prompt_tokens)

    # Uncomment for easy prompt token pressure
    # self.max_prompt_tokens = 1500
    # self.soft_max_prompt_tokens = 1000

    self.summaries = []
    self.recent_conversation = self.build_messages(self.prologue)

    if state.get('initialized'):
      self.load(state)

  def get_title(self):
    return self.persona_name

  def save(self):
    return {
      'initialized': True,
      'summaries': self.summaries,
      'recent_conversation': self.recent_conversation,
      'seed': self.seed,
      'persona_name': self.persona_name,
    }

  def load(self, state):
    self.summaries = state['summaries']
    self.recent_conversation = state['recent_conversation']
    self.seed = state.get('seed', self.seed)
    self.persona_name = state.get('persona_name', self.persona_name)

  def ask(self, raw_query):
    (query, self.response_prefix) = self.parse_query(raw_query)

    client_message = self.add_message(text=query, source='client')
    server_message = self.build_message(source='server')

    response = self.continue_conversation()
    for data in response:
      server_message['text'] += data
      yield data

    server_message['text'] = server_message['text'].strip()
    self.add_message(message=server_message)

    self.response_prefix = ''

  def continue_conversation(self):
    self.compress_conversation(self.max_response_tokens)

    conversation_prompt = self.format_conversation_prompt(self.recent_conversation)
    results = self.complete(
      conversation_prompt,
      max_length=self.max_response_tokens,
      stream=True
    )

    for i, data in enumerate(results):
      if self.response_prefix and i == 0:
        yield self.response_prefix
      yield data

  def build_message(self, text='', source=''):
    message = {
      'id': str(uuid.uuid4()),
      'text': text,
      'source': source,
    }
    return message

  def build_messages(self, messages):
    return [
      self.build_message(text=msg['text'], source=msg['source'])
      for msg in messages
    ]

  def add_message(self, message=None, text='', source=''):
    if not message:
      message = self.build_message(text=text, source=source)
    self.recent_conversation += [ message ]
    return message

  def delete_message(self, message={}, message_id=None):
    if not message_id:
      message_id = message['id']
    self.recent_conversation = [ m for m in self.recent_conversation if m['id'] != message_id ]

  def rollback(self):
    # TODO : do more stuff
    message_id = self.recent_conversation[-1]['id']
    self.delete_message(message_id=message_id)

  def shrink_messages(self):
    for (i, message) in enumerate(self.recent_conversation[:-8]):
      if not self.has_prompt_token_pressure():
        break
      if self.count_tokens(message['text']) < self.soft_max_message_tokens:
        continue

      response = self.complete(self.format_message_summary_prompt(message))
      message['text'] = f'summary=[ I {response.strip()} ]'

  def compress_conversation(self, space_required):
    self.shrink_messages()

    while self.get_conversation_space() <= space_required or len(self.recent_conversation) > self.soft_max_depth:
      if not self.has_prompt_token_pressure():
        break

      chunk = []
      chunk_text = ''
      while self.count_tokens(chunk_text) < self.min_rollup_tokens and len(self.recent_conversation) > 2:
        chunk += self.recent_conversation[:1]
        self.recent_conversation = self.recent_conversation[1:]
        chunk_text = '\n'.join([ m['text'] for m in chunk ])

      if len(chunk) == 0:
        breakpoint()
        break

      summary_prompt = self.format_summary_prompt(chunk)
      response = self.complete(summary_prompt)
      summary = f'{self.human_name} ' + response.strip()

      self.summaries = self.summaries[-(self.max_summaries - 1):] + [ summary ]

  def complete(self, prompt, max_length=None, stream=False):
    return self.llm.complete(
      prompt,
      model=self.model,
      temperature=float(self.temperature),
      max_length=max_length or self.max_response_tokens or 1000,
      stops=self.get_stops(),
      soft_stops=[ self.line_sep_token ],
      stream=stream,
    )

  def get_stops(self):
    return [ f'{self.human_name}>', f'{self.persona_name}>' ]

  def get_prompt_size(self):
    conversation_prompt = self.format_conversation_prompt(self.recent_conversation)
    return self.count_tokens(conversation_prompt)

  def get_conversation_space(self):
    return self.max_prompt_tokens - self.get_prompt_size() - self.max_response_tokens

  def has_prompt_token_pressure(self):
    return self.count_tokens(self.format_conversation_prompt(self.recent_conversation)) > self.soft_max_prompt_tokens;

  def format_summaries(self, include_pinned=True):
    pinned = [ self.pinned_summary() ] if include_pinned else []
    return [ f'- {s}' for s in pinned + self.summaries ]

  def format_message_summary_prompt(self, message, whitespace='\n'):
    author = self.persona_name if message["source"] == "server" else self.human_name
    return whitespace.join([
      self.format_message(message),
      f'{self.authority_name}> Summarize the previous message. Should be terse but informative. Use the past tense.',
      f'{author}> I',
    ])

  def format_summary_prompt(self, messages, whitespace='\n'):
    return whitespace.join([
      self.conversation_header(),
      self.format_messages(messages),
      f'{self.authority_name}> Give me your personal notes on the conversation. Should be terse but informative.',
      f'{self.persona_name}> {self.human_name}'
    ])

  def parse_query(self, query=''):
    parts = query.split(':>')
    if len(parts) == 1:
      return (query, '')
    return (' '.join(parts[:-1]), parts[-1].strip())

  def format_conversation_prompt(self, messages, whitespace='\n'):
    lines = [
      self.format_persona_bio(),
      self.summary_header(),
      *self.format_summaries(),
      self.conversation_header(),
      self.format_messages(messages),
      f'{self.format_seed()}{self.persona_name}>{self.response_prefix}'
    ]
    return whitespace.join(lines).strip()

  def count_tokens(self, text):
    return self.llm.count_tokens(text)

  def format_seed(self):
    if not self.seed:
      return ''
    return f'(In the following message, remember {self.seed})\n'

  def format_persona_bio(self):
    if not self.persona_bio:
      return ''
    return f'{self.persona_name} Bio: {self.persona_bio}'

  def format_messages(self, messages, wrap=True):
    lines = f'\n{self.line_sep}\n'.join(
      [ self.format_message(m) for m in messages ]
    )
    if wrap:
      lines = '\n'.join([ self.line_sep, lines, self.line_sep ])
    return lines

  def format_message(self, message):
    if message['source'] == 'server':
      return f'{self.persona_name}> {message["text"]}'
    else:
      return f'{self.human_name}> {message["text"]}'

  def inspect(self):
    return self.format_conversation_prompt(self.recent_conversation)

  def stats(self):
    history_size = len(self.recent_conversation)
    headroom = 100 - round(100 * self.get_prompt_size() / self.max_prompt_tokens)
    return f'( depth={history_size}, free={headroom}%, profile={self.profile} )'

  def capacity(self):
    return round(100 * self.get_prompt_size() / self.max_prompt_tokens)

  def variables(self):
    return [
      'model',
      'persona',
      'seed',
      'temperature',
      'max_response_tokens',
    ]

  @property
  def persona(self):
    return f'{self.persona_name} [ {self.persona_bio} ]'

  @persona.setter
  def persona(self, val):
    pattern = r"\s*([^\[]*)\s*(\[\s*([^\]]*)\s*\])?"
    match = re.match(pattern, val)
    if not match:
      return
    self.persona_name = match.group(1).strip()
    if match.group(3) != None:
      self.persona_bio = match.group(3).strip()



# npm install chatgpt puppeteer

import os, subprocess, base64, json

class ChatGPT:

  def __init__(self, conversation_id='', parent_message_id=''):
    self.conversation_id = conversation_id
    self.parent_message_id = parent_message_id
    self.spawn_node()

  def ask(self, prompt):
    result = self.send_message({
      'prompt': prompt,
      'conversation_id': self.conversation_id,
      'parent_message_id': self.parent_message_id,
    })
    self.conversation_id = result['conversation_id'];
    self.parent_message_id = result['message_id'];
    return result['response']

  def spawn_node(self):
    (self.read_pipe, self.write_pipe) = os.pipe()
    self.node_process = subprocess.Popen(
      [
        "node",
        "api.mjs"
      ],
      stdin=self.read_pipe,
      stdout=subprocess.PIPE,
      preexec_fn=os.setsid,
    )

  def send_message(self, msg):
    encoded_msg = base64.b64encode(json.dumps(msg).encode('utf-8'))

    os.write(self.write_pipe, encoded_msg + b'\n')
    response = self.node_process.stdout.readline()

    return json.loads(base64.b64decode(response).decode('utf-8'))

def set_subprocess_group():
    os.setpgid(0, 0)



import os, re, pyperclip

from ...utils import printer, open_in_editor, copy_to_clipboard
from ...publish import PublishGPT

from .command_manager import Commands
from .base_command import BaseCommand

@Commands.register('.clear')
class ClearCommand(BaseCommand):

  aliases = [ '.cl' ]
  help = 'Clears the terminal screen.'
  shortcut = [ 'C-x', 'C-c' ]
  erase_input = True

  def run(self):
    printer.clear(1)
    self.action('prompt')

@Commands.register('.copy')
class CopyCommand(BaseCommand):

  aliases = [ '.cp' ]
  help = 'Copies the first code block in the most recent response to the clipboard.'

  def run(self):
    self.banner()

    code_block = self.extract_code_block(self.repl.thread.messages)
    if code_block == None:
      return 'No code block found'

    success = copy_to_clipboard(code_block)
    if success:
      return 'Copied the first code block in the most recent response to the clipboard.'
    else:
      return 'Failed to copy a code block to the clipboard.'

  def extract_code_block(self, history):
    responses = [ entry.text for entry in history if entry.source == 'them' ]
    if len(responses) == 0:
      return None
    last_response = responses[-1]

    # This regex is too naive but is easy and usually works
    code_block_pattern = re.compile(r"```[a-zA-Z0-9]*?\n(.*?)```", re.DOTALL)
    code_blocks = code_block_pattern.findall(last_response)

    if len(code_blocks) == 0:
      return None

    return code_blocks[0]

@Commands.register('.debug')
class DebugCommand(BaseCommand):

  aliases = [ '.db' ]
  help = 'Triggers a breakpoint in the mode backend.'

  def run(self):
    self.repl.mode.debug()
    return ''

@Commands.register('.edit')
class EditCommand(BaseCommand):

  aliases = [ '.e' ]
  help = 'Opens the current message in the text editor specified by $EDITOR.'
  # We will use the built-in C-x+C-e keybinding
  # shortcut = [ 'C-x', 'C-e' ]
  erase_input = True

  def run(self):
    text = open_in_editor(self.repl.prompt.session, extension='.md')
    self.repl.auto_fills += [ text ]
    self.action('prompt')

@Commands.register('.exit')
class ExitCommand(BaseCommand):

  aliases = [ '.ex' ]
  help = 'Closes the REPL.'

  def run(self):
    raise KeyboardInterrupt

@Commands.register('.print')
class PrintCommand(BaseCommand):

  aliases = [ '.pp' ]
  help = 'Prints the internal prompt representation'
  shortcut = [ 'C-x', 'C-p' ]

  def run(self):
    self.banner()
    printer.print(self.repl.mode.inspect())
    return ''

# @Commands.register('.publish')
# class PublishCommand(BaseCommand):

#   help = 'Publishes the current thread online and prints the URL.'
#   aliases = [ '.pub' ]

#   def run(self):
#     self.banner()
#     publisher = PublishGPT(self.repl.thread)
#     return publisher.publish()

@Commands.register('.redo')
class RedoCommand(BaseCommand):

  aliases = [ '.r' ]
  help = 'Resubmits the most recent successful response.'
  shortcut = [ 'C-x', 'C-r' ]

  def run(self):
    self.banner()

    history = self.repl.thread.messages
    if len(history) < 2:
      return 'No message to redo'

    self.repl.mode.rollback()
    self.repl.mode.rollback()
    last_message = history[-2].text
    self.repl.thread.rollback(2)

    self.set_text(last_message)
    self.action('continue')

@Commands.register('.undo')
class UndoCommand(BaseCommand):

  aliases = [ '.u' ]
  help = 'Rolls back the most recent question and answer'
  shortcut = [ 'C-x', 'C-u' ]

  def run(self):
    self.banner()

    history = self.repl.thread.messages
    if len(history) < 2:
      return 'No message to undo'

    self.repl.mode.rollback()
    self.repl.mode.rollback()
    last_message = history[-2].text
    self.repl.thread.rollback(2)

    self.repl.auto_fills += [ last_message ]
    self.action('break')

@Commands.register('.rename')
class RenameCommand(BaseCommand):

  aliases = [ '.rnm' ]
  help = 'Renames the current thread. The new name must be provided as an argument to this command, e.g. ".rename my_new_name". The old thread isn\'t deleted.'

  def run(self):
    self.banner()
    if len(self.cmd_args) == 1:
      self.repl.thread.set_name(self.cmd_args[0])
      return f'Thread renamed to "{self.repl.thread.name}"'
    else:
      return 'Invalid thread name'

# I would prefer to call it ".reset", but .redo is more important
@Commands.register('.new')
class NewCommand(BaseCommand):

  aliases = [ '.n', '.reset' ]
  help = 'Resets the thread.'
  shortcut = [ 'C-x', 'C-n' ]

  def run(self):
    self.banner()
    self.repl.reset()
    printer.clear(2)
    return f'@{self.repl.thread.name} reset.'

@Commands.register('.seed')
class SeedCommand(BaseCommand):

  aliases = [ '.s' ]
  help = 'The text following .seed will be set as the conversation seed. Set no text to clear the seed.'

  def run(self):
    self.banner()
    self.repl.mode.set_seed(self.cmd_suffix)
    return 'Seed updated.'


@Commands.register('.get')
class GetCommand(BaseCommand):

  aliases = [ '.?' ]

  def run(self):
    self.banner()
    if len(self.cmd_args) == 0:
      variables = '\n'.join(
        sorted([ f'- {var}' for var in self.repl.mode.variables() ])
      )
      printer.print_markdown(f'Available mode variables:\n{variables}')
    else:
      key = self.cmd_args[0]
      val = getattr(self.repl.mode, key)
      if val == None or val == '':
        printer.print_markdown(f'**{key}** is empty')
      else:
        printer.print_markdown(f'**{key}** == **{val}**')
    return ''

@Commands.register('.set')
class SetCommand(BaseCommand):

  aliases = [ '.=' ]

  def run(self):
    self.banner()
    if len(self.cmd_args) == 0:
      variables = '\n'.join(
        sorted([ f'- {var}' for var in self.repl.mode.variables() ])
      )
      printer.print_markdown(f'Available mode variables:\n{variables}')
    else:
      key = self.cmd_args[0]
      val = ' '.join(self.cmd_args[1:])
      setattr(self.repl.mode, key, val)
      printer.print_markdown(f'**{key}** updated')
    return ''



import os
from .thread import Thread

class ThreadManager:

  def __init__(self, config):
    self.config = config

  def load(self, thread_name):
    thread = Thread(thread_name, self.config)
    thread.save()
    return thread

  def list(self):
    threads = []
    for file_name in os.listdir(self.config.threads_dir_path):
      if not file_name.endswith('.json'):
        continue
      thread_name = Thread.normalize_name(file_name[:-5])
      file_path = os.path.join(self.config.threads_dir_path, file_name)
      mod_time = os.path.getmtime(file_path)
      threads += [ (mod_time, thread_name) ]

    threads.sort(reverse=True)

    return [name for _, name in threads]

  def make_name(self):
    threads = self.list()
    i = len(threads) + 1
    while True:
      thread = f'thread-{i}'
      i += 1
      if thread not in threads:
        return thread



import os, re, pyperclip
from itertools import chain

def peek(gen):
  first = next(gen)
  return ( chain([ first ], gen), first )

class set_env:

  def __init__(self, name, value):
    self.name = name
    self.value = value

  def __enter__(self):
    self.old_value = os.environ.get(self.name)
    os.environ[self.name] = self.value

  def __exit__(self, *args):
    if self.old_value == None:
      del os.environ[self.name]
    else:
      os.environ[self.name] = self.old_value

def clear_screen():
  os.system('cls' if os.name == 'nt' else 'clear')

def copy_to_clipboard(text):
  try:
    pyperclip.copy(text)
    return True
  except Exception as e:
    return False



import traceback

# Rich Colors:
#  some= https://rich.readthedocs.io/en/stable/appendix/colors.html
#  all=  https://github.com/Textualize/rich/blob/master/rich/color.py
from rich.console import Console
from rich.live import Live
from rich.text import Text
from rich.table import Table

from .markdown import SmartMarkdown, GuessLexer

def render_markdown_to_html(text, title=''):
  svg = SmartMarkdown(f" \n{text}\n ", code_theme="monokai").to_svg(title=title)

  # The laziest hack ever
  svg = svg.replace('<g transform="translate(26,22)"', '<g transform="translate(-100,-100)"');
  svg = svg.replace('<rect fill="#292929"', '<rect fill="#000d1a"');

  return f"<div style=\"resize: horizontal; overflow: hidden; width: 1000px; height: auto;\">{svg}</div>"

class TempLog:

  def __init__(self, text=""):
    self.text = text

  def __enter__(self):
    print(f"\r{self.text}", end="")

  def __exit__(self, *args):
      print("\r" + " " * len(self.text), end="")
      print("\r", end="")

class RichLive:

  def __init__(self, console, transient=False):
    self.console = console
    self.transient = transient

  def update(self, content):
    max_height = self.console.height - 1
    raw_text = '\n'.join(content.split('\n')[-max_height:])
    text = Text.from_ansi(raw_text)

    self.live_enter.update(text, refresh=True)

  def __enter__(self):
    self.live = Live(
      console=self.console,
      vertical_overflow="vertical",
      auto_refresh=False,
      transient=self.transient,
    )
    self.live_enter = self.live.__enter__()
    self.live_enter._redirect_stdout=False

    return self

  def __exit__(self, *args):
    self.live.__exit__(*args)

class Printer:

  def __init__(self):
    self.console = Console(color_system="truecolor")

  def warmup(self):
    GuessLexer.warmup()
    # GuessLexer.load()

  def print(self, text):
    self.console.print(text)

  def to_markdown(self, text, preserve_softbreak=True, code_theme='monokai'):
    if preserve_softbreak:
      text = '  \n'.join(text.split('\n'))
    return SmartMarkdown(text, code_theme=code_theme)

  def print_markdown(self, *args, **kwargs):
    self.console.print(self.to_markdown(*args, **kwargs))

  def build_text(self, text_parts):
    text = Text()
    [ text.append_text(part) for part in text_parts ]
    return text

  def print_banner(self, bg_color='rgb(0,95,135)', text='', prefix='', title='', suffix=''):
    left_text = []
    if prefix:
      left_text += [ Text(prefix, style=f'grey70 on grey11') ]
    left_text += [ Text(text, style=f'bold white') ]

    center_text = []
    if title:
      center_text += [ Text(text, style=f'white') ]

    right_text = []
    if suffix:
      right_text += [ Text(suffix, style=f'rgb(200,200,200)') ]

    table = Table(
      expand=True,
      show_header=False,
      box=None,
      padding=0,
      row_styles=[ f'white on {bg_color}' ]
    )

    table.add_column("", justify="left")
    table.add_column("", justify="center")
    table.add_column("", justify="right")
    table.add_row(
      self.build_text(left_text),
      self.build_text(center_text),
      self.build_text(right_text),
    )

    self.console.print(table)

  def live(self, transient=False):
    return RichLive(self.console, transient=transient)

  def temp_log(self, text):
    return TempLog(text)

  def exception(self, e):
    self.print(f'\n[red]{"".join(traceback.TracebackException.from_exception(e).format())}[/red]')

  def pad_down(self, n=1):
    with printer.live(transient=True) as screen:
      screen.update('\n' * n)

  def clear(self, space=0):
    self.pad_down(self.console.height - space)

printer = Printer()



import uuid, re, os, json
from datetime import datetime
from .message import Message

class Thread:

  id = ''
  name = ''
  messages = []
  mode_name = ''
  mode_state = {}
  profile = ''
  seed = ''
  # timestamp = None

  def __init__(self, name, config):
    self.set_name(name)
    self.config = config

    if not self.load():
      self.reset()

  @property
  def mode(self):
    return Mode(self.mode_name, self.mode_state)

  def set_mode(self, mode_name, state=None):
    self.mode_name = mode_name
    self.mode_state = state or self.mode_state

  def get_file_path(self):
    file_name = f'{self.escape_name(self.name)}.json'
    return os.path.join(self.config.threads_dir_path, file_name)

  def to_data(self):
    return {
      'id': self.id,
      'name': self.name,
      'messages': [ msg.to_data() for msg in self.messages ],
      'mode_name': self.mode_name,
      'mode_state': self.mode_state,
      'profile': self.profile,
      'seed': self.seed,
      # 'timestamp': self.timestamp,
    }

  def load_data(self, data):
    self.id = data.get('id')
    self.name = data.get('name')
    self.messages = [
      Message().load_data(msg_data)
      for msg_data in data.get('messages', [])
    ]
    self.mode_name = data.get('mode_name')
    self.mode_state = data.get('mode_state')
    self.profile = data.get('profile')
    self.seed = data.get('seed')
    # self.timestamp = data.get('timestamp')
    return self

  def save(self):
    file_path = self.get_file_path()
    with open(file_path, 'w') as thread_file:
      json.dump(self.to_data(), thread_file, indent=2)

  def load(self):
    file_path = self.get_file_path()
    if not os.path.isfile(file_path):
      return False

    with open(file_path, 'r') as thread_file:
      self.load_data(json.load(thread_file))
    return True

  def set_name(self, name):
    self.name = self.normalize_name(name)

  def reset(self, preserve_profile=False, preserve_seed=False):
    self.id = self.id or str(uuid.uuid4())
    self.messages = []
    self.mode_name = self.mode_name or ''
    self.mode_state = {}
    self.profile = self.profile if preserve_profile else ''
    self.seed = self.seed if preserve_seed else ''
    # self.timestamp = datetime.now()

  def get_messages(self):
    return self.messages

  def add_message(self, source, text, stats=''):
    message = Message(source, text, stats=stats)
    self.messages += [ message ]
    return message

  def rollback(self, n=1):
    self.messages = self.messages[:-n]

  @classmethod
  def normalize_name(cls, thread_name):
    thread_name = thread_name.replace('@', '').replace('_', '-').strip()
    return re.sub(r'[^-a-zA-Z0-9\.]+', '_', thread_name)

  @classmethod
  def escape_name(cls, thread_name):
    return cls.normalize_name(thread_name).replace('-', '_')

# Just for dottable access
class Mode:
  def __init__(self, name, state):
    self.name = name
    self.state = state



from .base_mode import BaseMode, register_mode

@register_mode('echo')
class EchoMode(BaseMode):

  title = 'Bat'
  visible = False

  def ask(self, text):
    yield f'# {text}'

  def stats(self):
    return '( location="chamber" )'



from .base_mode import BaseMode, register_mode
from ..llms.chat_gpt import ChatGPT

@register_mode('chat-gpt')
class RawGPTMode(BaseMode):

  title = 'ChatGPT'
  loader_latency=.25

  def __init__(self, state={}):
    self.chat_gpt = ChatGPT(
      conversation_id=state.get('conversation_id', ''),
      parent_message_id=state.get('parent_message_id', ''),
    )

  def ask(self, text):
    return self.chat_gpt.ask(text)

  def save(self):
    return {
      'conversation_id': self.chat_gpt.conversation_id,
      'parent_message_id': self.chat_gpt.parent_message_id,
    }



import os, openai, re, sys, tiktoken

class GPT3:

  def __init__(self):
    if not os.environ.get('OPENAI_API_KEY'):
      print("Please set the OPENAI_API_KEY environment variable. If you don't have one you can generate one here https://beta.openai.com/account/api-keys")
      sys.exit(0)

    self.tokenizer = tiktoken.get_encoding("gpt2")

  def count_tokens(self, text):
    return len(self.tokenizer.encode(text))

  def complete(self, *args, **kwargs):
    if kwargs.get('stream', False):
      return self.__complete_async(*args, **kwargs)
    else:
      return self.__complete_sync(*args, **kwargs)

  def __complete_sync(self, *args, **kwargs):
    kwargs['stream'] = False
    response = self.get_response(*args, **kwargs)
    return response.choices[0].text

  def __complete_async(self, *args, **kwargs):
    kwargs['stream'] = True
    response = self.get_response(*args, **kwargs)
    for data in response:
      yield data.choices[0].text

  def get_response(self,
      prompt,
      max_length=1000,
      temperature=0.7,
      model="text-davinci-003",
      stops=None,
      soft_stops=[],
      stream=False,
    ):

    logit_bias = {}
    for s in soft_stops:
      logit_bias[s] = -100

    return openai.Completion.create(
      engine=model,
      prompt=prompt,
      max_tokens=max_length,
      temperature=temperature,
      top_p=1,
      frequency_penalty=0,
      presence_penalty=0,
      stop=stops,
      logit_bias=logit_bias,
      stream=stream,
      # logit_bias={
      #   '2949': 1, # Make "No" a more common token
      #   '44651': 1, # Make "Invalid" a more common token
      # },
    )



import uuid
from datetime import datetime

class Message:

  id = ''
  # timestamp = None
  source = ''
  text = ''
  stats = ''

  def __init__(self, source='', text='', stats=''):
    self.id = self.id or str(uuid.uuid4())
    # self.timestamp = datetime.now()
    self.source = source
    self.text = text
    self.stats = stats

  def to_data(self):
    return {
      'id': self.id,
      'source': self.source,
      # 'timestamp': self.timestamp,
      'text': self.text,
      'stats': self.stats,
    }
  def load_data(self, data):
    self.id = data.get('id')
    # self.timestamp = data.get('timestamp')
    self.source = data.get('source')
    self.text = data.get('text')
    self.stats = data.get('stats')
    return self



from .base_mode import BaseMode, register_mode

@register_mode('cowsay')
class CowsayMode(BaseMode):

  title = 'Cow'

  def ask(self, text):
    cow = f'```text\n{self.build_cow(text)}\n```'
    yield cow

  def stats(self):
    return '( mood=moootastic )'

  def inspect(self):
    return 'mooooooooooooooooooooooooooooooooooooooooooooooooooo'

  def build_cow(self, text):
    dialogue = f'< {text} >'
    border = f' { "-" * (len(dialogue) - 2) }'
    cow = f"""
{border}
{dialogue}
{border}
        \   ^__^
         \  (oo)\_______
            (__)\       )\/\\
                ||----w |
                ||     ||
"""
    return cow



from .base_mode import BaseMode, register_mode, get_mode, list_modes
from .synth_chat import SynthChatMode
from .echo import EchoMode
from .cowsay import CowsayMode
from .raw_gpt import RawGPTMode
from .raw_codex import RawCodexMode
from .chat_gpt import ChatGPT



from ...utils import printer

class BaseCommand:

  aliases = []
  help = ''
  erase_input = False
  sort_index = 1
  shortcut = []

  def __init__(self, repl, text, print_text):
    self.repl = repl
    self.text = text
    self.print_text = print_text
    self.banner_printed = False
    self.action_state = 'break'

    self.parse(text)

  def parse(self, raw_cmd):
    self.raw_cmd = raw_cmd.strip()
    parts = self.raw_cmd.split(' ')
    self.cmd_prefix = parts[0]
    self.cmd_args = parts[1:]
    self.cmd_suffix = ' '.join(self.cmd_args)

  def matches(self):
    return self.cmd_prefix == self.name or self.cmd_prefix in self.aliases

  def exec(self):
    output = None

    try:
      if not self.erase_input:
        self.print_text(self.text)
      output = self.run()
    except KeyboardInterrupt as e:
      raise e
    except Exception as e:
      if not self.banner_printed:
        self.banner()
      printer.exception(e)
      print()

    if output == '':
      print()
    elif output != None:
      print(output + '\n')

    return (self.action_state, self.text)

  def action(self, action_state):
    self.action_state = action_state

  def get_name(self):
    return self.__class__.name

  def set_text(self, text):
    self.text = text

  def banner(self, name='', newline=True):
    self.banner_printed = True
    if not name:
      name = self.get_name()
    if newline:
      print()
    printer.print_banner(
      bg_color='dark_red',
      text=f'[ {name} ] =>',
    )



import sys, re

from .pretty import PrettyPrintREPL
from .prompt import Prompt

from ..utils import peek, printer, Loader
from ..config import Config

from ..modes import get_mode
from .commands import Commands

class REPL:

  def __init__(
      self,
      thread_name=None,
      mode_name = None,
      auto_fills=[],
    ):
    self.pretty = PrettyPrintREPL(self)

    self.config = Config()
    self.config.load_plugins()

    self.thread = self.config.threads().load(thread_name)
    self.thread.set_mode(self.thread.mode.name or mode_name or 'synth-chat')

    self.mode_name = self.thread.mode.name
    self.auto_fills = auto_fills
    self.first_run = True

  def get_user_input(self):
    default = ''
    if len(self.auto_fills) > 0:
      default = self.auto_fills.pop(0)
    text = self.input(default=default)

    (action, new_text) = Commands.exec(
      repl=self,
      text=text,
      print_text=lambda s: self.pretty.print(s)
    )

    if action == 'prompt':
      return self.get_user_input()

    if action == 'break':
      return None

    if action == 'continue':
      return new_text

  def run(self):
    if len(self.thread.messages) == 0:
      printer.clear(1)
    else:
      self.warmup_thread()

    self.load_mode(self.mode_name, state=self.thread.mode.state, seed=self.thread.seed)
    self.create_prompt()

    self.pretty.intro()
    self.pretty.replay_thread()

    try:
      while True:
        self.core_loop()
    except (Exception, KeyboardInterrupt, EOFError) as error:
      self.pretty.leaving_thread()
      if isinstance(error, KeyboardInterrupt) or isinstance(error, EOFError):
        self.save_thread()
        sys.exit(0) # breaking might be better, but sys.exit is a lot faster
      else:
        printer.exception(error)
        sys.exit(1)

  def core_loop(self):
    self.pretty.your_banner(len(self.thread.messages) + 1, space=3)

    text = self.get_user_input()
    if text == None:
      return

    stats = self.mode.stats()
    self.pretty.their_banner(len(self.thread.messages) + 2, stats=stats, space=3)

    try:
      answer = self.ask(text)
    except (KeyboardInterrupt, EOFError):
      self.mode.rollback()
      self.pretty.request_canceled()
      return

    self.pretty.print(answer, newline=True)

    self.thread.add_message('you', text)
    self.thread.add_message('them', answer, stats=stats)

    self.save_thread()
    self.first_run = False

  def create_prompt(self):
    self.prompt = Prompt(self.config)
    self.prompt.bind_keys()
    Commands.bind_keys(self.prompt.kb)

  def input(self, default=''):
    seed = self.mode.get_seed()
    if seed:
      toolbar = f'seed={seed}'
    else:
      toolbar = 'No conversation seed set'
    return self.prompt.input(default=default, toolbar=toolbar)

  def save_thread(self):
    self.thread.set_mode(self.mode_name, self.mode.save())
    self.thread.seed = self.mode.get_seed()
    self.thread.save()

  def load_mode(self, mode_name, state={}, seed=''):
    self.mode = get_mode(mode_name)(state=state)
    self.mode.set_seed(seed)

  def reset(self):
    old_mode = self.mode

    self.load_mode(self.mode_name, seed=old_mode.get_seed())
    self.thread.reset()
    self.save_thread()

  def ask(self, text):
    delay = 0.25 if self.first_run else self.mode.loader_latency
    with Loader(show_timer=True, delay=delay):
      gen = iter(self.mode.ask(text))
      printer.warmup()
      response = peek(gen)[0]

    answer = ''
    with printer.live(transient=True) as screen:
      for data in response:
        answer += data
        display_text = self.pretty.partial_response(answer)
        screen.update(display_text)

    return answer.strip()

  # When replaying the thread, it might have to load guesslang
  # and the UI will seem to be frozen. This avoids that. It
  # usually won't run because SynthChat is pretty good about
  # labeling code blocks. Kinda jank since it's not actually
  # parsing the Markdown.
  def warmup_thread(self):
    messages = [ entry.text for entry in self.thread.messages ] + self.auto_fills
    pattern1 = re.compile(r'^```.+$', re.MULTILINE)
    pattern2 = re.compile(r'^```$', re.MULTILINE)

    should_warmup = any([
      len(re.findall(pattern1, msg)) != len(re.findall(pattern2, msg))
      for msg in messages
    ])
    if not should_warmup:
      return

    printer.clear(1)
    with self.pretty.loading_thread():
      printer.warmup()



from ..utils import printer

class PrettyPrintREPL:

  def __init__(self, repl):
    self.repl = repl

  def print(self, text, newline=False):
    printer.print_markdown(text)
    if newline:
      printer.print('')

  def intro(self):
    printer.print('Type "help" for a list of commands. Use "Enter" to submit and "Tab" to start a new line.\n')

  def your_banner(self, count, space=0):
    printer.print_banner(
      bg_color='rgb(0,95,135)',
      text=' You:',
      prefix=f' {count} ',
      suffix=f' @{self.repl.thread.name} [ {self.repl.mode_name} ]'
    )
    if space > 0:
      printer.pad_down(space)

  def their_banner(self, count, stats='', space=0):
    printer.print_banner(
      bg_color='spring_green4',
      text=f' {self.repl.mode.get_title()}:',
      prefix=f' {count} ',
      suffix=stats
    )
    if space > 0:
      printer.pad_down(space)

  def partial_response(self, text):
    # It would be great to use the monokai theme but it causes too much flashing
    markdown = printer.to_markdown(text.lstrip() + '█', code_theme='default')
    return markdown.to_text() + '\n\n\n'

  def loading_thread(self):
    return printer.temp_log(f'Loading \x1b[1m@{self.repl.thread.name}\x1b[0m...')

  def leaving_thread(self):
    print(f'\nLeaving \x1b[1m@{self.repl.thread.name}\x1b[0m')

  def request_canceled(self):
    printer.print('Request canceled\n')

  def replay_thread(self):
    for i, msg in enumerate(self.repl.thread.messages):
      if msg.source == 'you':
        self.your_banner(i + 1)
      elif msg.source == 'them':
        self.their_banner(i + 1, stats=msg.stats)
      self.print(msg.text, newline=True)



# Adapted from https://github.com/rawandahmad698/PyChatGPT/blob/master/src/pychatgpt/classes/spinner.py

from itertools import cycle
import threading
import time, sys, math

parts = '▁ ▂ ▃ ▄ ▅ ▆ ▅ ▄ ▃ ▁'.split(' ')
parts = [ '▁' ] * len(parts) + parts
frames = [
  ' '.join((parts[offset:] + parts[:offset])[:int(len(parts) / 2)])
  for offset in reversed(range(len(parts)))
]

class Loader:

    def __init__(self, show_timer=False, delay=0):
        self.show_timer = show_timer
        self.delay = delay
        self.running = False

        self.__screen_lock = threading.Event()
        self.__spinner = cycle(frames)
        self.__stop_event = False
        self.__thread = None

    def __enter__(self):
      self.start()
      return self

    def __exit__(self, *args):
      self.stop()

    def start(self):
        self.running = True
        start_time = time.time()
        self.__stop_event = False

        def run_spinner():
            while not self.__stop_event:
                time_diff = time.time() - start_time
                if time_diff > self.delay:
                    frame = next(self.__spinner)
                    if self.show_timer:
                        print(f"\r[ {math.floor(time_diff)}s ] {frame}  ", end="")
                    else:
                        print(f"\r{frame}  ", end="")
                time.sleep(0.05)

            self.__screen_lock.set()

        self.__thread = threading.Thread(target=run_spinner, args=(), daemon=True)
        self.__thread.start()

    def stop(self):
        if not self.running:
            return
        self.__stop_event = True
        if self.__screen_lock.is_set():
            self.__screen_lock.wait()
            self.__screen_lock.clear()
        print("\r                                              ", end="")
        print("\r", end="")



# INSTALLATION:
#   Copy or symlink this file into `$GPT_REPL_CONFIG_PATH/plugins/`
#   By default, this will be `~/.config/gpt_repl/plugins/`

from gpt_repl.modes import BaseMode, register_mode
from gpt_repl.llms.gpt3 import GPT3

@register_mode('bruh')
class BruhMode(BaseMode):

  title = 'Bruh'

  def __init__(self, state={}):
    self.llm = GPT3()

  def ask(self, text):
    prompt = f'Rephrase the following text to bruh-speak:\n{text}\n\nBruh:\n'
    return self.llm.complete(
      prompt,
      stream=True
    )



import importlib, re
from xml.dom import minidom

from rich.console import Console
from rich.markdown import Markdown, Paragraph, Heading, CodeBlock, BlockQuote, HorizontalRule, ListElement, ListItem, ImageItem, MarkdownElement, TextElement
from rich.syntax import Syntax

# The guesslang library loads quite slowly because of its dependency on TensorFlow,
# so we load it lazily using `importlib`
class GuessLexer:
  guess = None

  @classmethod
  def parse(cls, code):
    if code.strip() == '':
      return 'text'
    try:
      GuessLexer.load()
      return GuessLexer.guess._language_map[GuessLexer.guess.language_name(code.strip())]
    except Exception as e:
      print(e)
      return 'text'

  @classmethod
  def load(cls):
    if GuessLexer.guess == None:
      GuessLexer.guess = importlib.import_module('guesslang').Guess()

  @classmethod
  def warmup(cls):
    cls.load()
    GuessLexer.guess.language_name('_')


class SmartCodeBlock(CodeBlock):

  @classmethod
  def create(cls, markdown, node):
    node_info = node.info or ""
    lexer_name = node_info.partition(" ")[0]

    if not lexer_name and markdown.code_theme != 'default':
      lexer_name = GuessLexer.parse(str(node.literal))

    return cls(lexer_name or "text", markdown.code_theme)

  def __rich_console__(self, console, options):
    code = str(self.text).rstrip()
    if code.split('\n')[0].strip() != '':
      code = '\n' + code + '\n'
    else:
      code = code + '\n'

    if self.theme == 'default':
      yield code
    else:
      yield Syntax(
        code,
        self.lexer_name,
        theme=self.theme,
        word_wrap=True,
        padding=0
      )

class DocumentElement(TextElement):

  @classmethod
  def create(cls, markdown, node):
    return cls(markdown.code_theme)

  def __init__(self, theme):
    self.theme = theme

  def __rich_console__(self, console, options):
    if self.text.markup != "":
      yield ''
      code_block = SmartCodeBlock(lexer_name='html', theme=self.theme)
      code_block.text = self.format_html(self.text.markup)
      yield code_block

  def format_html(self, text):
    try:
      xml = minidom.parseString(text)
      lines = xml.toprettyxml(
        indent='  ',
      )[23:-1].split('\n')
      return '\n'.join([ line for line in lines if line.strip() != '' ])
    except Exception as e:
      return text


class SmartMarkdown(Markdown):
  elements  = {
    "paragraph": Paragraph,
    "heading": Heading,
    "code_block": SmartCodeBlock,
    "block_quote": BlockQuote,
    "thematic_break": HorizontalRule,
    "list": ListElement,
    "item": ListItem,
    "image": ImageItem,
    "document": DocumentElement,
  }

  def to_text(self, styles=True):
    console = Console(record=True)
    with console.capture() as capture:
      console.print(self)
    return console.export_text(styles=styles)

  def to_html(self):
    console = Console(record=True)
    with console.capture() as capture:
      console.print(self)
    return console.export_html(inline_styles=True)

  def to_svg(self, title=''):
    console = Console(record=True)
    with console.capture() as capture:
      console.print(self)
    return console.export_svg(title=title)



from .base_mode import BaseMode, register_mode
from ..llms.gpt3 import GPT3

@register_mode('raw-gpt')
class RawGPTMode(BaseMode):

  title = 'GPT'

  def __init__(self, state={}):
    self.llm = GPT3()

  def ask(self, text):
    return self.llm.complete(
      text,
      stream=True
    )



from prompt_toolkit import PromptSession
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.history import FileHistory
from prompt_toolkit.application.current import get_app
from prompt_toolkit import filters as Filters

class Prompt:

  def __init__(self, config):
    self.config = config
    self.start_session()

  def start_session(self):
    self.session = PromptSession(
      erase_when_done=True,
      history=FileHistory(self.config.prompt_history_path),
    )
    self.kb = KeyBindings()

  def bind_keys(self):
    @self.kb.add('tab', filter=is_not_searching)
    def _(event):
      prefix = event.current_buffer.document.leading_whitespace_in_current_line
      event.current_buffer.insert_text('\n' + prefix)

    @self.kb.add('enter', filter=is_not_searching)
    def _(event):
      if len(event.current_buffer.text.strip()) > 0:
        event.current_buffer.validate_and_handle()
      else:
        event.current_buffer.insert_text('\n')

  def input(self, default='', toolbar='', extension='.md'):
    text = self.session.prompt(
      '',
      multiline=True,
      key_bindings=self.kb,
      enable_open_in_editor=True,
      tempfile_suffix=extension,
      default=default,
      bottom_toolbar=toolbar,
    )
    return text.strip()

@Filters.Condition
def is_not_searching():
  return not get_app().layout.is_searching



from ...utils import printer

from .command_manager import Commands
from .base_command import BaseCommand

@Commands.register('help')
class HelpCommand(BaseCommand):

  aliases = [ '.help', 'help()' ]
  help = 'Prints a list of available commands and a brief description of each.'

  def matches(self):
    return self.raw_cmd == self.name or self.raw_cmd in self.aliases

  def run(self):
    self.banner()
    printer.print_markdown(self.format_help())
    return ''

  def format_help(self):
    return f"""
{self.commands()}

{self.shortcuts()}

{self.tips()}
""".strip()

  def commands(self):
    cmd_lines = [
      f'**{cmd.name}** or {cmd.aliases[0]}: {cmd.help}'
      for cmd in Commands.registry.values()
      if cmd.help
    ]
    formatted_lines = '\n'.join(sorted(cmd_lines))
    return f"""
## Commands
{formatted_lines}
""".strip()

  def shortcuts(self):
    cmd_lines = [
      f'**{" + ".join(cmd.shortcut)}**: {cmd.help}'
      for cmd in Commands.registry.values()
      if cmd.help and len(cmd.shortcut) > 0
    ]
    formatted_lines = '\n'.join(sorted([
      *cmd_lines,
      '**C-c**: Closes the REPL.',
      '**C-d**: Closes the REPL.',
      '**C-r**: Search message history.',
      '**C-x + C-e**: Opens the current message in the text editor specified by $EDITOR.',
    ]))
    return f"""
## Shortcuts
**Enter**: Submits the current message.
**Tab**: Adds a new line.
{formatted_lines}
""".strip()

  def tips(self):
    return """
## Tips:
- To shape your conversation, consider setting a conversation seed using the `.seed DESCRIPTION` command. Some examples:
    - `.seed You must use Markdown headers on every message`
    - `.seed You must contantly use exclamation marks`
    - `.seed You must speak in rhymes`
  Seeds are best phrased as commands or statements of fact.
- If you don't like the most recent response, regenerate it with `.redo`
- If the thread has gone completely off the rails, reset it with `.new`
 - Many commands have shorter aliases e.g. `.seed` has `.s`. See the command list for more info.
""".strip()

@Commands.register('.invalid')
class InvalidCommand(BaseCommand):

  # This ensures this command runs last
  sort_index = 2

  def matches(self):
    return self.raw_cmd.startswith('.')

  def run(self, *args):
    self.banner()
    return f'Command "{self.raw_cmd}" is invalid. Use .help for a list of valid commands.'



import click, os, datetime
from os.path import abspath

from .repl import REPL
from .config import Config
from .modes import list_modes

@click.command(help='Usually just `gpt-repl @thread-name [-m MODE]`')
@click.argument("cmd", type=str, required=False)
@click.option('-m', '--mode', default='synth-chat', help="REPL mode to load")
@click.option('-t', '--thread', default=None, help="Thread name to open")
# @click.option('-f', '--file', type=click.Path(exists=True), help="Path to text file to preload")
# @click.option('-l', '--list', is_flag=True, help="List all threads")
def run(cmd, mode, thread):
  config = Config()

  if str(cmd) == 'help':
    ctx = click.get_current_context()
    click.echo(ctx.get_help())
    return

  if cmd == 'threads':
    for thread in config.threads().list():
      print(thread)
    return

  if str(cmd)[0] == '@':
    thread = cmd[1:]

  if str(cmd) == '-':
    thread = config.threads().list()[0]

  if thread == None:
    thread = config.threads().make_name()

  repl = REPL(
    mode_name=mode,
    thread_name=thread,
  )

  if cmd == 'modes':
    for mode in list_modes():
      print(mode)
    return

  repl.run()

if __name__ == '__main__':
  run()



import typer

app = typer.Typer()


@app.command()
def hello(name: str):
    print(f"Hello {name}")


@app.command()
def goodbye(name: str, formal: bool = False):
    if formal:
        print(f"Goodbye Ms. {name}. Have a good day.")
    else:
        print(f"Bye {name}!")


if __name__ == "__main__":
    app()



from .base_mode import BaseMode, register_mode
from ..llms.gpt3 import GPT3

@register_mode('raw-codex')
class RawCodexMode(BaseMode):

  title = 'Codex'

  def __init__(self, state={}):
    self.llm = GPT3(model='code-davinci-002', temperature=0.1)

  def ask(self, text):
    yield '```\n'
    for data in self.llm.complete(text, stream=True):
      yield data
    yield '\n```'



import os, json, re, pdb, importlib.util
from os import path, listdir
from .utils import printer
from .threads import ThreadManager

class Config:

  default_config_path = '~/.config/gpt_repl'

  def __init__(self, config_path=None):
    config_path = config_path or os.getenv('GPT_REPL_CONFIG_PATH', self.default_config_path)

    self.config_dir_path = path.abspath(path.expanduser(config_path))
    self.config_file_path = path.join(self.config_dir_path, 'config.json')
    self.prompt_history_path = path.join(self.config_dir_path, 'prompt_history')
    self.threads_dir_path = path.join(self.config_dir_path, 'threads')
    self.plugins_dir_path = path.join(self.config_dir_path, 'plugins')

    dirs = [ self.config_dir_path, self.threads_dir_path, self.plugins_dir_path]
    for dir in dirs:
      if not path.exists(dir):
        os.makedirs(dir)

    self.reload()

  def save(self):
      with open(self.config_file_path, 'w') as config_file:
        json.dump(self.config or {}, config_file, indent=2)

  def reload(self):
    self.config = {}

    if not path.exists(self.config_file_path):
      self.save()

    with open(self.config_file_path, 'r') as config_file:
      self.config = json.load(config_file)

    return self.config

  def threads(self):
    return ThreadManager(self)

  def load_plugins(self):
    files = os.listdir(self.plugins_dir_path)
    for file in files:
      file_path = path.join(self.plugins_dir_path, file)
      if not file_path.endswith('.py'):
        continue
      try:
        spec = importlib.util.spec_from_file_location('plugins', file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
      except Exception as e:
        printer.exception(e)
        continue



mode_registry = {}

def register_mode(name):
  def decorator(cls):
    mode_registry[name] = cls
    return cls
  return decorator

def get_mode(name):
  return mode_registry[name]

def list_modes(show_hidden=False):
  modes = []
  for name, mode in mode_registry.items():
    if name != 'base' and mode.visible or show_hidden:
      modes += [ name ]
  modes.sort()
  return modes

@register_mode('base')
class BaseMode:

  title = ''
  visible = True
  loader_latency=1.5

  seed = ''

  def __init__(self, state={}):
    pass

  def get_title(self):
    return self.title

  def ask(self, text):
    yield ''

  def get_seed(self):
    return self.seed

  def set_seed(self, seed):
    self.seed = seed

  def rollback(self, message_id=''):
    pass

  def save(self):
    return {}

  def stats(self):
    return ''

  def inspect(self):
    return ''

  # To use ipdb set:
  #   export PYTHONBREAKPOINT=IPython.terminal.debugger.set_trace
  def debug(self):
    breakpoint()

  def variables(self):
    return []
